/*
	Copyright (C) 2011 Micha Hersch, University of Lausanne, Swiss Institute of Bioinformatics

	This program is free software: you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation, either version 3 of the License, or
	any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/


/**
 * \file snake.h
 * \author Micha Hersch
 * \date 27-02-10
 */

#ifndef __SNAKE_H__
#define __SNAKE_H__

#include <iostream>
#include "ip_public.h"


#define printVec2(a) cout<<a.x<<" "<<a.y;


#define MAX_SNAKE_SIZE 200
#define SNAKE_EYECUP

#include "openCotyledon.h"

typedef OpenCotyledon cot_t;




/** Hypocotyl tracking snakes. Like the ones used for contour detection
 * but here snakes are open and tend to grow up. 
 */

class Snake {

 protected:
  Vec2 points[MAX_SNAKE_SIZE];//the set of points defining the snake
  int size;
  int offset[2];
  int hneb; //half the neighboorhood size
  int neb_size;
  CvMat *Eint;
  CvMat *Eext;
  CvMat *Etot;
  int anchored;
  
  int valid;
  float Eint_weight;

  cot_t cot;

  OpenCotyledon base;


 protected: //methods
  /** print the energy landscape, useful for debugging */
  void printEnergies();

 
 public:  //methods

  Snake(int h=5);
  ~Snake();

  /** Initializes the snake. Initialization is done with Canny edge detection
      \param img the image
      \param nb_points the number of points in the snake
      \param rect a region of interest in the image where the hypocotyl
      is expected
      \param canny parameter of the canny edge detection
      \param ground the ground level, from which the hypocotyl is expected to
      grow (if available)
  */
  int init(IplImage *img,int nb_points, CvRect *rect=NULL,float canny=0.9,int ground=0,int threshold=0);
  /** Sets the half size of the neighbourhood in which to minimize the energy */
  void setNeighborhoodSize(int hs);
  /** Adds a point to the snake */
  int addPoint(Vec2 p)
  {if(size<MAX_SNAKE_SIZE){points[size++]=p;return size;}return 0;};
  int addPoint()
  {if(size<MAX_SNAKE_SIZE){points[size]=points[size-1];return ++size;}return 0;};
  int addPoints(int n){for(int i=0;i<n;i++){if(!addPoint())break;}return size;};
  int removePoints(int nb);
  /** empty the snake */
  void empty(){size=0;}
  void wakeUp(int n){valid=1;size=n;} 
  int isValid()const{return size && valid;}
  void validate(){valid=1;}
  void invalidate(){valid=0;}
  /** find the closest point in the snake.
      \param pos the point for which we want to be close
      \param squaredist the squared euclidean distance to the closest
      point in snake
      \return the index of the closest point in snake
  */
  int findClosestPoint(Vec2 pos, int *squaredist);
  /** 
     \return the number of points in the snake
   */
  int getSize()const{return size;}

  /** Sets the ratio between internal and external energy.
      This value is actually a mean value, there is some noise added to it
      to get out of local minima
      \param Eint_weight the energy ratio
  */
  void setEnergyBalance(float b){Eint_weight=b;}
  /** Gets the ratio between internal and external energy
      \sa setEnergyBalance(float b)
   */
 float getEnergyBalance()const{return Eint_weight;}

 /** Draws the snake on the tracked hypocotyl.
     \param disp the image where to draw it
     \param col the color in which to draw it
  */
  void draw(IplImage *disp, CvScalar col){
    Vec2 os(0,0);//offset
    if(disp->roi){os=Vec2(disp->roi->xOffset,disp->roi->yOffset);}
    for(u8 i=0;i<size-1;i++){
      cvLine(disp,(points[i]-os).to2d(),(points[i+1]-os).to2d(),col,3);
    }
  }
   

  void drawLength(IplImage *disp,float len, CvScalar col);

  /** Updates the snake.
      \param img the image containing the hypocotyl
      \param nbit the number of iterations for the update
      \param eyecup no tip update if different from 0, no base update if <0
   */
  void updateSnake(IplImage *img,int nbit,int eyecup=0);

 
  /** Updates the tip snake, only the top most points
      \param img the image containing the hypocotyl
      \param nbit the number of iterations for the update
      \param n the number of points to update
  */
  void updateTip(IplImage *img,int nbit, int n);


  void downFromTip(IplImage *img,float angle,float length);

  /** Force the lower end to be at ground level. It also removes points below
      ground level
      \param ground the ground level (in pixels)
   */
  void anchorToGround(int ground);
  void drop(int x,int y,int size,int spread);


  /** \return a reference to the cotyledon
   */
  cot_t& getCotyledon(){return cot;};
  OpenCotyledon& getBase(){return base;};
  /** Draws the cotyledon
      \param img where to draw it
      \param ox x-offset (if the img has a ROI)
      \param oy y-offset (if the img has a ROI)
   */
  void drawCotyledon(IplImage *img,int ox=0,int oy=0){
    if(img->roi){ox=img->roi->xOffset;oy=img->roi->yOffset;}
    cot.draw(img,ox,oy);}

  void drawBase(IplImage *img,int ox=0,int oy=0)
  {if(img->roi){ox=img->roi->xOffset;oy=img->roi->yOffset;}
    base.draw(img,ox,oy,CV_RGB(0,0,255));}

  /** Updates the cotyledon.
      \param img the image
      \param img_old previous img, same size as img
  */
  void updateCotyledon(IplImage *img, IplImage *img_old){cot.update(img,img_old);}

  
  Vec2 updateBase(IplImage *img, IplImage *img_old)
  {return base.update(img,img_old);}
  
/** Set the position of a point
   \param i the point to move
   \param p where to move it
  */
  void setPoint(int i,Vec2 p){if(i<size)points[i]=p;}
  /** 
   * Sets the position of the last point (the bottom)
   * @param p new position
   */
  void setLastPoint(Vec2 p){points[size-1]=p;}
/** Get the positions of a point. Zero'th position can always be queried
   \param i the point to query
   \return its position
  */

  Vec2 getLastPoint()const{return points[size-1];}

/** Get the positions of a point. Zero'th position can always be queried
   \param i the point to query
   \return its position
  */

  Vec2 getPoint(int i)const
  {if((!i) || i<size)return points[i];std::cerr<<"out of bound"<<std::endl;return Vec2();}
  void updatePoint(int i, IplImage *img);
  void updateInternalEnergy(int i);
  void updateExternalEnergy(int i, IplImage *img);
  // max x coordinate
  int maxX(){int mx=0;for(u8 i=0;i<size;i++)mx=max(mx,points[i].x);return mx;}
 // max y coordinate
  int maxY(){int my=0;for(u8 i=0;i<size;i++)my=max(my,points[i].y);return my;}
 // min x coordinate
  int minX(){int mx=points[0].x;for(u8 i=1;i<size;i++)mx=min(mx,points[i].x);return mx;}
 // min y coordinate
  int minY(){int my=points[0].y;for(u8 i=1;i<size;i++)my=min(my,points[i].y);return my;}
  
  void printPoints(std::ostream& out)const;

  void mark(IplImage *img,int i)const;
  /**
   * \param len the length over which  the final orientation is averaged
   * \return final with upward vertical in degrees (counterclockwise)
   */
  float getOrientation(float len)const;

  /**
   * \return total length in pixels,
   */
  float getLength()const;

  void translate(Vec2 trans){for(u8 i=0;i<size;i++){points[i] += trans;}
    cot.translate(trans);base.translate(trans);}

  
  void translatePoint(u16 i,Vec2 trans)
  {if(i<size)points[i] += trans;}
  //  void finalOrientation

};

int findMinVertDist(CvSeq *seq, int y0);

static std::ostream& operator<<(std::ostream& out, const Snake& s)
  {s.printPoints(out);  return out;}


/* Sort 2d points in top-to-bottom left-to-right order */
static int cmp_func( const void* _a, const void* _b, void* userdata )
{
    CvPoint* a = (CvPoint*)_a;
    CvPoint* b = (CvPoint*)_b;
    int y_diff = a->y - b->y;
    int x_diff = a->x - b->x;
    return y_diff ? y_diff : x_diff;
}





#endif
