/*
	Copyright (C) 2011 Micha Hersch, University of Lausanne, Swiss Institute of Bioinformatics

	This program is free software: you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation, either version 3 of the License, or
	any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

/**
 * \file snakeNest.h
 * \author Micha Hersch
 * \date 01-03-10
 */

#ifndef __SNAKE_NEST_H__
#define __SNAKE_NEST_H__

//#define WITH_EQUALIZATION
//#define WITH_GLOBAL_THRESHOLDING
#define WITH_LOCAL_THRESHOLDING



#ifdef WITH_EQUALIZATION
#define GLOBAL_PREPROCESSING
#endif
#ifdef WITH_GLOBAL_THRESHOLDING
#define GLOBAL_PREPROCESSING
#endif
#ifdef WITH_LOCAL_THRESHOLDING
#define GLOBAL_PREPROCESSING
#endif

#define WITH_HISTORY


#include "snake.h"
#include "imgGrabber.h"


#ifdef WITH_HISTORY
#include "nestHistory.h"
#endif


enum NestState{NONE,ZOOM,GET_GROUND,CORRECT,CALIB};

class SnakeNest;


/**
 * A set of many snakes for analyzing images containing multiple hypocotyls
 * 
 */
class SnakeNest{

 protected:
  Snake *snakes;
  int nb;
  int nrow;
  int ncol;

  int nb_points;

  int *grounds;


  ImgGrabber *grabber;
  IplImage *img; //img to be processed
  IplImage *old_img;

#ifdef GLOBAL_PREPROCESSING
  IplImage *imgeq; //img to be preprocessed
  IplImage *old_imgeq; //img to be preprocessed
#endif

  IplImage *disp; //image to display the snakes
  IplImage *disp_bak;// raw image backup
  IplImage *idisp; //image to display the interface

  char wname[80];
 

  int zooming;
  CvRect zoomRect;
  CvRect zoomRect_bak;
  float zoom_factor;
  float win_factor;

  int thresh_w;
  int thresh_h;

  int selected_snake;
  int selected_point;
  int more_selected_points;

  int moving_snake;
  Vec2 initial_snake_pos;

  int *bw_thresholds;//contains local thresholding values
  int global_threshold;

  int saveImages;
  int keep_trace;
  int move_base;

  bool oriwin;//1 window to set orientation length parameter is active

  char imgfileformat[5];
  char baseimgfile[80];

#ifdef WITH_HISTORY
  NestHistory history;  
#endif


  NestState state;
 public:
  static int modulation;
  static int max_modulation;
  static float max_balance;
  static int detect;
  static int max_detect;
  static SnakeNest *instance;
  static int bw_threshold; //black&white threshold for image binarization

  static int orient_length;
  static int max_orient_length;

 public:
  SnakeNest(int hor=1,int ver=1,int length=100);
  ~SnakeNest();
  void setNeighborhood(int half_neb)
  {for(int i=0;i<nb;i++)getSnake(i)->setNeighborhoodSize(half_neb);};
 
  /** 
   * initializes the image grabber
   * @param fname the name of the first image file
   */
  void initGrabber(const char *fname);

  /** 
   * grabs the next frame
   */
  int nextFrame();

  /** 
   * tries a first guesses where the hypocotyls are
   * 
   * @param quincunx if different than 0 expects the cotyledon to be arrange in
   * quincunx, with the first row shifted to the right
   */
  void initSnakes(int quincunx=0);
  void basicInitSnakes(int quincunx=0);
  /** 
   * Moves all the snakes of an given amount 
   * 
   * @param v the amount to move in pixels
   */
  void translateAll(Vec2 v){
    for(int i=0;i<nb;i++)
      {Snake* s=getSnake(i);if(s->isValid())s->translate(v);}};

  /** 
   * Updates all valid snakes
   * 
   */
  void updateSnakes(){
    for(int i=0;i<nb;i++)
      {Snake* s=getSnake(i);if(s->isValid())s->updateSnake(img,10);}};
  
  /** 
   * Updates the tip (top 10 points) of all valid snakes
   * 
   */
  void updateSnakeTips(){ 
    for(int i=0;i<nb;i++)
      {Snake* s=getSnake(i);if(s->isValid())s->updateTip(img,10,10);}};
  
/** 
   * Initializes the window display.
   * 
   * @param win_name the window name
   * @param autosize 1 of AUTOSIZE is true, 0 otherwise
   * @param trackbar 1 for a trackbar
   */
  void initDisplay(const char *win_name="window",int autosize=1,int trackbar=1);
  void closeDisplay();
  /** 
   * records the whole hypocotyl in a file
   * 
   */
  void keepTrace(){keep_trace=1;}
  void moveBase(){move_base=1;}
  
/** 
   * Opens outfiles files
   * 
   * @param pofname base file name
   * @param or_out  orientation stream
   * @param len_out length stream
   * @param seg_out initial segment stream
   * 
   * @return 1 if all files can be opened, 0 otherwise
   */
  int openResFiles(const char *pofname, std::ofstream& or_out,
		   std::ofstream& len_out, std::ofstream& seg_out, 
		   std::ofstream& tip_out);
  /** 
   * remove a set of snake from the nest
   * 
   * @param filename a file containing one number per line, the snake
   * to be discarded (index starts at 1)
   */
  void discardSnakes(const char *filename);
  void initCotyledons(const char *cotfname);
  void display(CvScalar col=CV_RGB(255,0,0));
  Snake *getSnake(int i){if(i<nb)return snakes+i;return NULL;}
  Snake *getSnake(int x,int y){int i=y*ncol+x;if(i<nb)return snakes+i;return NULL;}
  /**
   * getting reading access to snake
   * \param i snake index
   */
  const Snake *getSnake(int i)const{if(i<nb)return snakes+i;return NULL;}
  /**
   * getting reading access to snake
   */
  const Snake *getSnake(int x,int y)const{int i=y*ncol+x;if(i<nb)return snakes+i;return NULL;}

  /**
   * getting reading access to processed image img
   */
  const IplImage *getImage(){return img;}
  
  void retrieveBackground()
  {cvSetImageROI(img,zoomRect);cvSetImageROI(disp_bak,zoomRect); 
    cvCvtColor(disp_bak,img,CV_BGR2GRAY);cvResetImageROI(img);
    cvResetImageROI(disp_bak);}


  /** 
   * Retrieves the snake closest to a given position. Non valid snakes
   * (of size zero) are discarded 
   * 
   * @param v the reference position 
   * @param bot 1 if considering only the bottom of the snakes
   * @param exclude snake to be excluded from the competition
   * (-1 if no snake needs to be excluded,-2 if exclude all invalid snake)
   * 
   * @return the index of the closest snake
   */
  int closestSnake(Vec2 v,int bot=0,int exclude=-1);

  /** 
   * Retrieves the snake closest to a given position. Only non valid snakes
   * (of size zero) are considered 
   * 
   * @param v the reference position 
   * @param bot 1 if considering only the bottom of the snakes
   * @param exclude snake to be excluded from the competition
   * (-1 if no snake needs to be excluded)
   * 
   * @return the index of the closest snake
   */
  int closestSleepingSnake(Vec2 v);
  void setState(NestState s);
  NestState getState(){return state;};
  void setGround(int ground);
  /** 
   * Sets the ground level from gui
   * 
   */ 
 void getGround(int event,int x,int y){
    switch(event){
    case CV_EVENT_LBUTTONUP:setGround(y);break;
    case CV_EVENT_RBUTTONUP:wakeUpSnake(x,y);break;
    case CV_EVENT_MOUSEMOVE:
      if(RND(10)<1){
	cvCopy(disp,idisp);
	cvLine(idisp,cvPoint(0,y),cvPoint(idisp->width-1,y),CV_RGB(255,255,255));
	cvShowImage(wname,idisp);cvWaitKey(1);break;   
      }
    }
 }
 void correct(int event,int flags,int x,int y);

/**
     @brief main loop runs the processing for open cotyledon and high quality images
     @param neb_size neighborhood size for hypocotyl tracking
     @param imfname the first image file
     @param outfname the file where to write the snakes
   */
 void run(int neb_size,const char *imfname,const char *outfname);


  void zoom(Vec2 p, float factor);
  void zoom(int col,int row,int frame);
  void zoom();
  void resizeWindow(float fac);
 
/** 
    unzooms    
    @param per if 0 unzoom is temporary (for processing) otherwise it's permanent(also for display),  
*/
  void unzoom(int per=0);
  void resetZoom();
  void zoomThresholding(float z);
  void selectPoint(int x,int y);
  void selectSnake(int x,int y);
  void selectSnake(int i);
  void selectManyPoints(int x,int y);
  void wakeUpSnake(int x,int y)
  {int i=closestSleepingSnake(Vec2(x,y));
      if(i!=-1){getSnake(i)->wakeUp(nb_points);selectSnake(i);}}
  Snake *selectedSnake()
  {if(selected_snake!=-1){return getSnake(selected_snake);}return NULL;}

  void tuneSelectedSnake(float mod);
/* { */
/*    int ec = 0; */
/* #ifdef SNAKE_EYECUP */
/*     ec =EYECUP; */
/* #endif */
/*     Snake *s = selectedSnake();if(s){s->setEnergyBalance(mod);s->updateSnake(img,10,ec);}} */
  
  void resetSelectedPoint(int x,int y)
  {selected_point=moving_snake=more_selected_points=-1;display();};
 
  void inactivateCotyledon(int x,int y)
  {getSnake(closestSnake(Vec2(x,y)))->getCotyledon().inactivate();display();}
  
  int localThresholding(int i);
  void setBWThreshold(int thresh){bw_threshold=thresh;
    if(selected_snake>=0)bw_thresholds[selected_snake] = thresh;}

  void threshold(int i=-1){//TODO: we ca use a ROI here
    //   int j=(i>=0)?i:selected_snake;bw_threshold=bw_thresholds[j];
    if(i>=0){bw_threshold=bw_thresholds[i];}
    else{
      if(global_threshold==-1 && selected_snake !=-1)
	{bw_threshold=bw_thresholds[selected_snake];}
      else{
	global_threshold=bw_threshold;
	for(int i=0;i<nb;i++){bw_thresholds[i]=bw_threshold;}
      }
    }
    if(bw_threshold>-1){
      /* cvSetImageROI(img,zoomRect);//for later use */
      /* cvSetImageROI(old_img,zoomRect); */
      /* cvSetImageROI(imgeq,zoomRect);//for later use */
      /* cvSetImageROI(old_imgeq,zoomRect); */
      cvThreshold(old_imgeq,old_img,bw_threshold,255,CV_THRESH_BINARY);
      cvThreshold(imgeq,img,bw_threshold,255,CV_THRESH_BINARY);
    }
  }

  void bwUpdateSnake(int i, int nbit,int eyecup);

/** 
   * draws all the cotyledons
   * 
   */ 
  void drawCotyledons(){
    for(int i=0;i<nb;i++)getSnake(i)->drawCotyledon(disp);
  }

  void showOrientationLength();

  /** 
   * selects the closest snake for moving
   * 
   */  
  void selectMovingSnake(int x,int y)
  {int c=closestSnake(Vec2(x,y),0,-2);selectSnake(c);
    if(selected_snake==-1){selectSnake(x,y);}//to make sure it's selected
    initPointMove(); moving_snake=selected_snake;initial_snake_pos=Vec2(x,y);}
  
/* void moveSelectedSnake(int x, int y) */
/*   {if(moving_snake!=-1){ */
/*       getSnake(moving_snake)->translate(Vec2(x,y)-initial_snake_pos); */
/*       drawSingleSnake(moving_snake);}} */

/*   void resetMovingSnake(int x, int y) */
/*     {moveSelectedSnake(x,y);moving_snake=-1;} */

  void initPointMove(){cvCopy(disp_bak,disp);
    for(int i=0;i<nb;i++)if(i!=selected_snake && getSnake(i)->isValid()) getSnake(i)->draw(disp,CV_RGB(255,0,0));} 


  void moveSelectedPoint(int x,int y)
  {
    if(moving_snake==-1 && selected_point==-1 )return;// for more security
    if(moving_snake==selected_snake){
      getSnake(moving_snake)->translate(Vec2(x,y)-initial_snake_pos);
      initial_snake_pos = Vec2(x,y);
    }
    else{
      if(selected_point!=-1){
	getSnake(selected_snake)->setPoint(selected_point,Vec2(x,y));
	if(more_selected_points!=-1){
	  for(int i=selected_point;i>more_selected_points;i--){
	    getSnake(selected_snake)->setPoint(i-1,Vec2(x,y));
	  }
	}
      }
    }
    drawSingleSnake(selected_snake);
  };

  void drawSingleSnake(int i){
     cvCopy(disp,idisp);getSnake(selected_snake)->draw(idisp,CV_RGB(255,0,0));
     cvShowImage(wname,idisp);}

  void readCotyledons(const char *fname);
  void writeCotyledons(const char *fname);
  void printSnakes(std::ostream& out)const;
  int printSnakes(const char *basename, int i)const;
  void printOrientations(std::ostream& out,float length);
  void printSegments(std::ostream& out);
  void printLengths(std::ostream& out);
  void validateAll(){for(int i=0;i<nb;i++)getSnake(i)->validate();}
  void invalidateAll(){for(int i=0;i<nb;i++)getSnake(i)->invalidate();}
  void saveImage(int n);
  void markCotyledons();
  void getThresholdValues(int i=-1);
  void bwUpdateSnake(int i);
  void waitForCommand(int x,int y);
  /** 
   * sets the base name for the filename where to save the images
   * 
   * @param outimgfile eg: image.tif (to get image001.tif, image002.tiff,...
   * 
   * @return 1 if successful, 0 if can't extract the file name extension
   */
  int setImageFileName(char *outimgfile);


#ifdef WITH_HISTORY
 void saveSnakes(uint tim);
 void backToTime(uint tim);
 bool presentTime()const
 {return grabber->GetCurrentTimeIndex()-3==history.getPresent();}
#endif


};

int SnakeNest::detect = 10;
int SnakeNest::max_detect=20;

int SnakeNest::modulation = 20;
int SnakeNest::max_modulation=50;
float SnakeNest::max_balance=1;

int SnakeNest::bw_threshold= 110;

int SnakeNest::orient_length=15;
int SnakeNest::max_orient_length=100;

SnakeNest *SnakeNest::instance = NULL;



static void on_mouse(int event,int x,int y,int flags, void *params){
  SnakeNest *nest = (SnakeNest *)params;
 /*  if(flags & CV_EVENT_FLAG_ALTKEY){//moving many points */
/* 	    std::cout<<"malt"<<std::endl; */
/*  } */
/*   Snake *s; */
/*   int c; */
  switch(nest->getState()){
  case GET_GROUND: // setting the ground position
    nest->getGround(event,x,y);
    break;
  case CORRECT:
    nest->correct(event,flags,x,y);
    break;
  default:;
  }
}

static void on_mouse_thresholding(int event,int x,int y,int flags, void *params){
  int z = 0;
  if(event== CV_EVENT_RBUTTONDOWN){
    z = 1;
  }
  if(event== CV_EVENT_LBUTTONDOWN){
    z = -1;
  }
  SnakeNest::instance->zoomThresholding(z);
 

}
static void detect_tbar(int detect){
  SnakeNest::detect = max(1,detect);
  if(SnakeNest::instance){
    SnakeNest::instance->initSnakes();
    SnakeNest::instance->updateSnakes();
    SnakeNest::instance->display(CV_RGB(255,0,0));
  }
}


static void modulation_tbar(int modulation){
  SnakeNest::modulation=modulation;
  if(SnakeNest::instance){
    SnakeNest::instance->tuneSelectedSnake(( SnakeNest::max_balance*modulation)/SnakeNest::max_modulation);
      SnakeNest::instance->display(CV_RGB(255,0,0));
    }
}

static void threshold_tbar(int thresh){
  SnakeNest::instance->setBWThreshold(thresh);
  SnakeNest::instance->threshold();
  cvShowImage("thresholding",SnakeNest::instance->getImage());
}


static void orient_tbar(int len){
  SnakeNest::orient_length = len;
  SnakeNest::instance->display();
  SnakeNest::instance->showOrientationLength();
}
      
#endif
